/*  $Id: sglr-interface.c 20451 2006-10-11 10:35:13Z jurgenv $  */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <aterm1.h>
#include <atb-tool.h>
#include <Error-utils.h>

#include "sglr.h"
#include "sglr-strings.h"
#include "mem-alloc.h"
#include "parser.h"
#include "forest.h"
#include "rsrc-usage.h"


ATerm  SG_TermToToolbus(ATerm t);
static size_t SG_FileSize(const char *FN);
void   SG_PrintToken(FILE *out, token c);
void   SG_Validate(const char *caller);
static char  *SG_ReadFile(const char *prg, const char *err, const char *fnam, size_t *fsize);

static FILE* stream_to_parse = NULL;

/*
 Controlling global behaviour: verbosity, debugging mode, dot
 output of parse forests/parse stacks, and so on.
 */

int  _SG_Mode = 0;
char *SG_Path;

/*
    |SGinitParser| initializes the parser's operating mode to a
    sensible default
 */

void  SGinitParser(ATbool toolbus_mode)
{
  if(toolbus_mode) {
    SG_TOOLBUS_ON();
  }

  SG_FILTER_ON();
  SG_FILTER_REMOVECYCLES_OFF();
  SG_FILTER_DIRECTEAGERNESS_ON();
  SG_FILTER_EAGERNESS_ON();
  SG_FILTER_INJECTIONCOUNT_ON();
  SG_FILTER_PRIORITY_ON();
  SG_FILTER_REJECT_ON();

  SG_BINARY_ON();
}

void SGshowMode()
{
  ATfprintf(stderr, "TOOLBUS:     %s\n", SG_TOOLBUS?"y":"n");
  ATfprintf(stderr, "ERROR:       %s\n", SG_ERROR?"y":"n");
  ATfprintf(stderr, "VERBOSE:     %s\n", SG_VERBOSE?"y":"n");
  ATfprintf(stderr, "DEBUG:       %s\n", SG_DEBUG?"y":"n");
  ATfprintf(stderr, "SHOWSTAT:    %s\n", SG_SHOWSTAT?"y":"n");
  ATfprintf(stderr, "ASFIX2ME:    %s\n", SG_ASFIX2ME?"y":"n");
  ATfprintf(stderr, "BINARY:      %s\n", SG_BINARY?"y":"n");
  ATfprintf(stderr, "SHOWSTACK:   %s\n", SG_SHOWSTACK?"y":"n");
  ATfprintf(stderr, "FILTER:      %s\n", SG_FILTER?"y":"n");
  ATfprintf(stderr, "STARTSYMBOL: %s\n", SG_STARTSYMBOL?"y":"n");
}

/**  The ToolBus API functions  **/

/*
 The function |SGopenLanguageFromTerm| initializes the parse table for
 language |L| from the term |tbl|.
 */

ATerm SGopenLanguageFromTerm(language L, ATerm tbl, const char *path)
{
  parse_table *pt = NULL;

  IF_STATISTICS(ATfprintf(SG_log(), "Language: %t\n", L); SG_Timer());

  pt = SG_BuildParseTable((ATermAppl) tbl, path);

  IF_STATISTICS(ATfprintf(SG_log(),
                          "Obtaining parse table for %t took %.6fs\n",
                          L, SG_Timer()));
  if (pt) {
    SG_SaveParseTable(L, pt);
  }

  return (ATerm) (pt ? ATempty : NULL);
}

ATerm SGopenLanguage(language L, const char *FN, const char *inFile)
{
  parse_table *pt;

  if (!L || !FN) {
    return (ATerm) NULL;
  }

  SG_Validate("SGopenLanguage");
  IF_VERBOSE(
    if (FN)
      ATwarning("opening parse table %s\n", SG_SAFE_STRING(FN))
  );

  pt = SG_AddParseTable(L, FN, inFile);

  return (ATerm) (pt ? ATempty : NULL);
}

label start_label;

ATerm SGparseStream(FILE *f,
		    SGLR_ParseTable parseTable,
		    const char *top,
		    const char *path,
		    int stop)
{
  int i;
  int maxProd = SG_PT_NUMPRODS((parse_table*)parseTable);

  forest ret;
  size_t ntok = 0;
  parse_table *pt;

  ATerm toptrm;

  stream_to_parse = f;

  start_label = -1;

  /* This is cache-able */
  if (stop) {
    toptrm = ATmake("<str>", top);

    for (i = 0; i <= maxProd; ++i) {
      ATerm nt;
      if (ATmatch((SG_PT_PRODUCTIONS(((parse_table*)parseTable)))[i], "prod([<term>], cf(sort(<term>)), <term>)", NULL, &nt, NULL)) {
	if (ATisEqual(toptrm, nt)) {
	  start_label = i+SG_PROD_START;
	  break ;
	}
      }
    }
  }

  SG_Validate("SGparseStream");

  ret = SG_Parse(path, (parse_table*)parseTable, top, SG_GetChar, 10000);

  return (ATerm) ret;
}

ATbool SGisParseTree(ATerm t)
{
  return ATgetAFun(t) == SG_ParseTree_AFun;
}

ATbool SGisParseError(ATerm t)
{
  return ATgetAFun(t) == SG_ParseError_AFun;
}

int SG_GetChar() {
  return getc(stream_to_parse);
}

void SG_UnGetChar() {
  fseek(stream_to_parse, -1L, SEEK_CUR);
}

#if 1
ATerm SG_TermToToolbus(ATerm t)
{
  SG_InitPTGlobals();  /*  Make REALLY sure the PT globals are initialised  */

  return SG_TOOLBUS
    ? (ATerm) ATmakeAppl1(SG_SndValue_AFun, t)
    : t;
}
#endif

FILE  *SG_Log = NULL;

FILE  *SG_OpenLog(const char *prg, const char *fnam)
{
  if (SG_Log)
    return SG_Log;

  if (!fnam || !strcmp(fnam, ""))
    fnam = ".sglr-log";
  if (!(SG_Log = fopen(fnam, "wb"))) {
    ATerror("%s%sCannot create logfile %s\n",  SG_COND_STRING(prg), prg?": ":"", fnam);
  }
  IF_VERBOSE(ATwarning("%s%slogging to %s\n",  SG_COND_STRING(prg), prg?": ":"", fnam));
  return SG_Log;
}

FILE  *SG_log(void)
{
  return SG_Log;
}

void SG_SetLog(FILE* stream)
{
  SG_Log = stream;
}

void SG_CloseLog(void)
{
  if(SG_Log) fclose(SG_Log);
}

FILE *SG_OpenFile(const char *std_error, const char *FN)
{
  FILE *file;

  if (!FN || !strcmp(FN, "") || !strcmp(FN, "-")) {
    if (!std_error || (FN && !strcmp(FN, "-"))) {
      return stdin;
    }

    return NULL;
  }

  if (!(file = fopen(FN, "rb"))) {
    return NULL;
  }
  return file;
}

void SG_CloseFile(FILE *fd)
{
  if(fd != stdin)
    fclose(fd);
}


static size_t SG_FileSize(const char *FN)
{
  struct stat  statbuf;

  if (!FN || !strcmp(FN, "") || !strcmp(FN, "-"))
    return -1;  /*  We can't tell how many tokens to read  */

  if(stat(FN, &statbuf) < 0) {
    return -1;  /*  We can't tell how many tokens to read, and
    may not be able to read anything... */
  }
  return statbuf.st_size;
}

/*
 |SG_PrintToken| prints token |c| on file |out|
 */

void SG_PrintToken(FILE *out, token c)
{
  int i = SG_GETTOKEN(c);

  if(i != SG_EOF &&isgraph(i)) {
    fprintf(out, "%c", i);
  } else {
    switch(i) {
      case SG_EOF:
        fprintf(out, "EOF");
        break;
      case '\t':
        fprintf(out, "\\t");
        break;
      case '\n':
        fprintf(out, "\\n");
        break;
      case '\r':
        fprintf(out, "\\r");
        break;
      default:
        fprintf(out, "\\%d", i);
        break;
    }
  }
}

/*
 SG_Validate: check the `mode' bits, to make sure that no
 debugging/statistics output is written unless the logging
 file is opened, &c.
 */

void SG_Validate(const char *caller)
{
  IF_STATISTICS(
    if (!SG_log()) {
      ATwarning("Warning: implicitly opening logfile\n");
      SG_OpenLog(caller, ".sglr-log");
    }
  );
}


